*  Real Inverse FFT
*****************************************************************************
*
* FILENAME      : ifft_rl.asm
*
* WRITTEN BY    : Daniel Mazzocco
*                 Texas Instruments, Houston
*
* DATE          : 18th Feb 1992
*
* VERSION       : 1.0
*
*****************************************************************************
* VER	  DATE		COMMENTS
* ---	------------	-----------------------------------------------------
* 1.0   18th Feb 92     Original Release. Started from forward real FFT
*                       routine written by Alex Tessarolo, rev 2.0 .
*
*****************************************************************************
*
* SYNOPSIS:     int     ifft_rl( FFT_SIZE, LOG_SIZE, SOURCE_ADDR, DEST_ADDR,
*				 SINE_TABLE, BIT_REVERSE );
*
*		int	FFT_SIZE	; 64, 128, 256, 512, 1024, ...
*		int	LOG_SIZE	;  6,   7,   8,   9,   10, ...
*               float   *SOURCE_ADDR    ; Points to where data is originated
*                                       ; and operated on.
*               float   *DEST_ADDR      ; Points to where data will be stored.
*		float	*SINE_TABLE	; Points to the SIN/COS table.
*		int	BIT_REVERSE	; =  0, Bit Reversing is disabled.
*					; <> 0, Bit Reversing is enabled.
*
*		NOTE:	1) If SOURCE_ADDR = DEST_ADDR, then in place bit
*			   reversing is performed, if enabled (more
*			   processor intensive).
*			2) FFT_SIZE must be >= 64 (this is not checked).
*
* DESCRIPTION:  Generic function to do an inverse radix-2 FFT computation
*               on the C30.
*               The data array is FFT_SIZE-long with real and imaginary
*		points R and I as follows:
*
*                       SOURCE_ADDR[0]          -> R(0)
*						   R(1)
*						   R(2)
*						   R(3)
*						   .
*					           .
*						   R(FFT_SIZE/2)
*					           I(FFT_SIZE/2 - 1)
*						   .
*					           .
*						   I(2)
*                     SOURCE_ADDR[FFT_SIZE - 1] -> I(1)
*
*               The output data array will contain only real values .
*               Bit reversal is optionally implemented at the end
*		of the function.
*
*		The sine/cosine table for the twiddle factors is expected
*		to be supplied in the following format:
*
*		SINE_TABLE[0]		   -> sin(0*2*pi/FFT_SIZE)
*				   	      sin(1*2*pi/FFT_SIZE)
*                                			.
*                                       		.
*					      sin((FFT_SIZE/2-2)*2*pi/FFT_SIZE)
*		SINE_TABLE[FFT_SIZE/2 - 1] -> sin((FFT_SIZE/2-1)*2*pi/FFT_SIZE)
*
*		NOTE: The table is the first half period of a sine wave.
*
*		Stack structure upon call:
*
*                        +-------------+
*		-FP(7)   | BIT_REVERSE |
*		-FP(6)   | SINE_TABLE  |
*		-FP(5)   | DEST_ADDR   |
*		-FP(4)   | SOURCE_ADDR |
*		-FP(3)   | LOG_SIZE    |
*		-FP(2)   | FFT_SIZE    |
*		-FP(1)   | return addr |
*		-FP(0)   | old FP      |
*			 +-------------+
*
*****************************************************************************
*
*		NOTE:	 Calling C program can be compiled using either large
*			 or small model.
*
*		WARNING: DP initialised only once in the program. Be wary
*			 with interrupt service routines. Make sure interrupt
*			 service routines save the DP pointer.
*
*               WARNING: The SOURCE_ADDR must be aligned such that the first
*			 LOG_SIZE bits are zero (this is not checked by the
*			 program).
*
*****************************************************************************
*
* REGISTERS USED: R0, R1, R2, R3, R4, R5, R6, R7
*		  AR0, AR1, AR2, AR3, AR4, AR5, AR6, AR7
*		  IR0, IR1
*		  RC, RS, RE
*		  DP
*
* MEMORY REQUIREMENTS:  Program = 322 Words (approximately)
*			Data    =   7 Words
*			Stack	=  12 Words
*
*****************************************************************************
*
* BENCHMARKS:	Assumptions 	- Program in RAM0
*				- Reserved data in RAM0
*				- Stack on Primary/Expansion Bus RAM
*				- Sine/Cosine tables in RAM0
*				- Processing and data destination in RAM1.
*				- Primary/Expansion Bus RAM, 0 wait state.
*
*		FFT Size 	Bit Reversing	Data Source	Cycles(C30)
*		--------	-------------	-----------	-----------
*               1024            OFF             RAM1            25892 approx.
*               Note: This number does not include the C callable overheads.
*                     Add 57 cycles for these overheads.
*****************************************************************************

FP		.set	AR3

                .global _ifft_rl         ; Entry execution point.

FFT_SIZE:       .usect  ".ifftdata",1    ; Reserve memory for arguments.
LOG_SIZE:       .usect  ".ifftdata",1
SOURCE_ADDR:    .usect  ".ifftdata",1
DEST_ADDR:      .usect  ".ifftdata",1
SINE_TABLE:     .usect  ".ifftdata",1
BIT_REVERSE:    .usect  ".ifftdata",1
SEPARATION:     .usect  ".ifftdata",1

;
; Initialise C Function.
;

                .sect   ".iffttext"

_ifft_rl:      PUSH    FP              ; Preserve C environment.
		LDI	SP,FP
		PUSH	R4
		PUSH	R5
                PUSH	R6
		PUSHF	R6
		PUSH	R7
		PUSHF	R7
		PUSH	AR4
		PUSH	AR5
		PUSH	AR6
		PUSH	AR7
		PUSH	DP

		LDP	FFT_SIZE	; Initialise DP pointer.

		LDI	*-FP(2),R0	; Move arguments from stack.
		STI	R0,@FFT_SIZE
		LDI	*-FP(3),R0
		STI	R0,@LOG_SIZE
		LDI	*-FP(4),R0
		STI	R0,@SOURCE_ADDR
		LDI	*-FP(5),R0
		STI	R0,@DEST_ADDR
		LDI	*-FP(6),R0
		STI	R0,@SINE_TABLE
		LDI	*-FP(7),R0
		STI	R0,@BIT_REVERSE


;
; Perform Last FFT loops first (loop 2 onwards).
;
;				 LOOP
;			       1st  2nd .....
;     _         ______________  \/  \/
;    |         |____X'(I1)____| 0   0  <-  X'(I1) + X'(I3)
;    |  AR1 -> |__X(I1)_(1st)_| 1   1  <-  X(I1) + X(I2)
;    |         |__X(I1)_(2nd)_| 2   2       .
;    |         |__x(I1)_(3rd)_| 3   3       .
;    |         |       .      |
;    |                 .
;    |    A -> |______________|
;    |         |____X'(I2)____| 8   16  <- X'(I2) * 2
;    |    B -> |       .      |
;    |                 .
;    |         |______________|
;    |	       |__X(I2)_(3rd)_| 13  29      .
;    |         |__X(I2)_(2nd)_| 14  30      .
;    |  AR2 -> |__X(I2)_(1st)_| 15  31 <-  X(I4) - X(I3)
;    |         |____X'(I3)____| 16  32 <-  X'(I1) - X'(I3)
;    |  AR3 -> |__X(I3)_(1st)_| 17  33 <-  [X(I1)-X(I2)]*COS-[X(I3)+X(I4)]*SIN
;    |         |__X(I3)_(2nd)_| 18  34      .
;    |         |__X(I3)_(3rd)_| 19  35      .
;    |         |       .      |
;    |                 .
;    |    C -> |______________|
;    |         |____X'(I4)____| 24  48 <- -X'(I4) * 2
;    |    D -> |       .      |
;    |                 .
;    |         |______________|
;    |         |__X(I4)_(3rd)_| 29  61      .
;    |         |__X(I4)_(2nd)_| 30  62      .
;    |_ AR4 -> |__X(I4)_(1st)_| 31  63 <-  [X(I1)-X(I2)]*SIN+[X(I3)+X(I4)]*COS
;              |______________| 32  64
;	AR1 -> |______________| 33  65
;	       |       .      |
;                      .
;                      .
;                     \|/
;

                LDI     1,IR0           ; step between two consecutive sinus
                LDI     4,R5            ; stage number from 4 to M.
                LDI     @FFT_SIZE,R7
                LSH     -2,R7           ; R7 is FFT_SIZE/4-1 (ie 15 for 64 pts)
                SUBI    1,R7            ; and will be used to point at A & D .
                LDI     @FFT_SIZE,R6    ; R6 will be used to point at D.
                LSH     1,R6
                LDI     @SOURCE_ADDR,AR5
                LDI     @SOURCE_ADDR,AR1

LOOP:           LSH     -1,R6           ; R6 is FFT_SIZE at the 1st loop
                LDI     AR1,AR4
		ADDI	R7,AR1		; AR1 points at A.
		LDI	AR1,AR2
		ADDI	2,AR2  		; AR2 points at B.
		ADDI	R6,AR4
		SUBI	R7,AR4		; AR4 points at D.
		LDI	AR4,AR3
		SUBI	2,AR3           ; AR3 points at C.

		LDI	R7,IR1
		LDI	R7,RC

INLOP:          ADDF3   *--AR1(IR1),*--AR3(IR1),R0  ; R0 = X'(I1) + X'(I3) ---+
                SUBF3   *AR3,*AR1,R1                ; R1 = X'(I1) - X'(I3) -+ |
                LDF     *--AR4,R2                   ;                       | |
             || STF     R0,*AR1++                   ; X'(I1) <--------------|-+
                MPYF    -2.0,R2                     ; R2 = -2*X'(I4) --+    |
                LDF     *--AR2,R3                   ;                  |    |
             || STF     R1,*AR3++                   ; X'(I3) <---------|----+
                MPYF    2.0,R3                      ; R3 = 2*X'(I2) -, |
                STF     R3,*AR2++(IR1)              ; X'(I2) <-------' |
             || STF     R2,*AR4++(IR1)              ; X'(I4) <---------+

                LDI     @FFT_SIZE,IR1 ; IR1=SEPARATION BETWEEN SIN/COS TBLS
		LDI	@SINE_TABLE,AR0 ; AR0 points at SIN/COS table.
                LSH     -2,IR1
                SUBI    3,RC

                SUBF3   *AR2,*AR1,R3            ; R3 = X(I1)-X(I2)
                ADDF3   *AR1,*AR2,R2            ; R2 = X(I1)+X(I2) ---+
                MPYF3   R3,*++AR0(IR0),R1       ; R1 = R3*SIN         |
                LDF     *AR4,R4                 ; R4 = X(I4)          |
                MPYF3   R3,*++AR0(IR1),R0       ; R0 = R3*COS         |
             || SUBF3   *AR3,R4,R3              ; R3 = X(I4)-X(I3)  --|--+
                ADDF3   R4,*AR3,R2              ; R2 = X(I3)+X(I4)    |  |
             || STF     R2,*AR1++               ; X(I1) <-------------+  |
                MPYF3   R2,*AR0--(IR1),R4       ; R4 = R2*COS            |
             || STF     R3,*AR2--               ; X(I2) <----------------+
                ADDF3   R4,R1,R3                ; R3 = R3*SIN + R2*COS -----+
                MPYF3   R2,*AR0,R1              ; R1 = R2*SIN               |
             || STF     R3,*AR4--               ; X(I4) <-------------------+
                SUBF3   R1,R0,R4                ; R4 = R3*COS - R2*SIN

                RPTB    IN_BLK

                SUBF3   *AR2,*AR1,R3            ; R3 = X(I1)-X(I2)
                ADDF3   *AR1,*AR2,R2            ; R2 = X(I1)+X(I2) ---+
                MPYF3   R3,*++AR0(IR0),R1       ; R1 = R3*SIN         |
             || STF     R4,*AR3++               ; X(I3)               |
                LDF     *AR4,R4                 ; R4 = X(I4)          |
                MPYF3   R3,*++AR0(IR1),R0       ; R0 = R3*COS         |
             || SUBF3   *AR3,R4,R3              ; R3 = X(I4)-X(I3)  --|--+
                ADDF3   R4,*AR3,R2              ; R2 = X(I3)+X(I4)    |  |
             || STF     R2,*AR1++               ; X(I1) <-------------+  |
                MPYF3   R2,*AR0--(IR1),R4       ; R4 = R2*COS            |
             || STF     R3,*AR2--               ; X(I2) <----------------+
                ADDF3   R4,R1,R3                ; R3 = R3*SIN + R2*COS -----+
                MPYF3   R2,*AR0,R1              ; R1 = R2*SIN               |
             || STF     R3,*AR4--               ; X(I4) <-------------------+
IN_BLK:         SUBF3   R1,R0,R4                ; R4 = R3*COS - R2*SIN

                SUBF3   *AR2,*AR1,R3            ; R3 = X(I1)-X(I2)
                ADDF3   *AR1,*AR2,R2            ; R2 = X(I1)+X(I2) ---+
                MPYF3   R3,*++AR0(IR0),R1       ; R1 = R3*SIN         |
             || STF     R4,*AR3++               ; X(I3)               |
                LDF     *AR4,R4                 ; R4 = X(I4)          |
                MPYF3   R3,*++AR0(IR1),R0       ; R0 = R3*COS         |
             || SUBF3   *AR3,R4,R3              ; R3 = X(I4)-X(I3)  --|--+
                ADDF3   R4,*AR3,R2              ; R2 = X(I3)+X(I4)    |  |
             || STF     R2,*AR1                 ; X(I1) <-------------+  |
                MPYF3   R2,*AR0--(IR1),R4       ; R4 = R2*COS            |
             || STF     R3,*AR2                 ; X(I2) <----------------+
                LDI     R6,IR1                  ; Get prepared for the next
                ADDF3   R4,R1,R3                ; R3 = R3*SIN + R2*COS -----+
                MPYF3   R2,*AR0,R1              ; R1 = R2*SIN               |
             || STF     R3,*AR4++(IR1)          ; X(I4) <-------------------+
                SUBF3   R1,R0,R4                ; R4 = R3*COS - R2*SIN
                NEGF    *AR1++(IR1),R2          ; DUMMY
             || STF     R4,*AR3++(IR1)          ; X(I3)               |

                SUBI3   AR5,AR1,R0
		CMPI	@FFT_SIZE,R0
		BLTD	INLOP			; LOOP BACK TO THE INNER LOOP
                NOP     *AR2++(IR1)             ; DUMMY
		LDI	R7,IR1
		LDI	R7,RC

		ADDI	1,R5
                CMPI    @LOG_SIZE,R5            ; next stage if any left
		BLED	LOOP
                LDI     @SOURCE_ADDR,AR1
                LSH     1,IR0                   ; double step in sinus table
                LSH     -1,R7

;
; Perform Third FFT loop .
;
;
; Part A:
;     _         ______
;    |  AR1 -> |__I1__| 0  <-  X(I1) + X(I3)
;    |	       |______| 1
;    |  AR2    |__I2__| 2  <-  2 * X(I2)
;    |	       |______| 3
;    |  AR3 -> |__I3__| 4  <-  X(I1) - X(I3)
;    |	       |______| 5
;    |  AR4 -> |__I4__| 6  <- -2 * X(I4)
;    |_	       |______| 7
;       AR1 -> |______| 8
;    	       |______| 9
;              |   .  |
;		   .
;                  .
;                 \|/

                LDI     @SOURCE_ADDR,AR1
		LDI	AR1,AR2
		LDI	AR1,AR3
		LDI	AR1,AR4
                ADDI    2,AR2
                ADDI    4,AR3
                ADDI    6,AR4
                LDI     8,IR0
		LDI	@FFT_SIZE,RC
                LSH     -3,RC
                SUBI    1,RC
		LDI	@SINE_TABLE,AR0 ; AR0 points at SIN/COS table.

                RPTB    LOOP3_A
                LDF     *AR3,R3
                ADDF3   R3,*AR1,R0                  ; R0 = X'(I1) + X'(I3) ---+
                SUBF3   R3,*AR1,R1                  ; R1 = X'(I1) - X'(I3) -+ |
                LDF     *AR4,R2                     ;                       | |
             || STF     R0,*AR1++(IR0)              ; X'(I1) <--------------|-+
                MPYF    -2.0,R2                     ; R2 = -2*X'(I4) --+    |
                LDF     *AR2,R3                     ;                  |    |
             || STF     R1,*AR3++(IR0)              ; X'(I3) <---------|----+
                MPYF    2.0,R3                      ; R3 = 2*X'(I2) -, |
LOOP3_A:        STF     R3,*AR2++(IR0)              ; X'(I2) <-------' |
             || STF     R2,*AR4++(IR0)              ; X'(I4) <---------+

;
; Part B:
;     _         ______
;    |         |______| 0
;    |  AR1 -> |__I1__| 1 <-  X(I1) + X(I2)
;    |         |______| 2
;    |  AR2 -> |__I2__| 3 <-  X(I4) - X(I3)
;    |         |______| 4
;    |  AR3 -> |__I3__| 5 <- [X(I1)-X(I2)]*COS-[X(I3)+X(I4)]*SIN
;    |         |______| 6
;    |_ AR4 -> |__I4__| 7 <- [X(I1)-X(I2)]*SIN+[X(I3)+X(I4)]*COS
;              |______| 8
;       AR1 -> |______| 9       NOTE: COS(2*pi/8) = SIN(2*pi/8)
;	       |   .  |
;                  .
;                  .
;                 \|/
;

                LDI     @SOURCE_ADDR,AR1
		LDI	AR1,AR2
		LDI	AR1,AR3
		LDI	AR1,AR4
                ADDI    1,AR1
                ADDI    3,AR2
                ADDI    5,AR3
                ADDI    7,AR4
                LDI     @SINE_TABLE,AR7 ; AR7 points at SIN/COS table.
		LDI	@FFT_SIZE,RC
                LSH     -3,RC
                LDI     RC,IR1
		SUBI	2,RC


                LDF     *AR2,R6                 ; R6 = X(I2)
                LDF     *AR3,R0                 ; R0 = X(I3)
                ADDF3   R6,*AR1,R5              ; R5 = X(I1)+X(I2) ---------+
                SUBF3   R6,*AR1,R4              ; R4 = X(I1)-X(I2)          |
                SUBF3   R0,R4,R3                ; R3 = X(I1)-X(I2)-X(I3)    |
                ADDF3   R0,R4,R2                ; R2 = X(I1)-X(I2)+X(I3)    |
                SUBF3   R0,*AR4,R1              ; R1 = X(I4)-X(I3) ---------|-+
             || STF     R5,*AR1++(IR0)          ; X(I1) <-------------------+ |
                ADDF3   R2,*AR4,R5              ; R5 = X(I1)-X(I2)+X(I3)+X(I4)|
             || STF     R1,*AR2++(IR0)          ; X(I2) <---------------------+
                MPYF3   R5,*++AR7(IR1),R1       ; R1 = R5*SIN    --------------+
             || SUBF3   *AR4,R3,R2              ; R2 = X(I1)-X(I2)-X(I3)-X(I4) |
                MPYF3   R2,*AR7,R0              ; R0 = R2*SIN    ---+          |
             || STF     R1,*AR4++(IR0)          ; X(I4) <-----------|----------+
                                                ;                   |
                RPTB    LOOP3_B                 ;                   |
                                                ;                   |
                LDF     *AR2,R6                 ; R6 = X(I2)        |
             || STF     R0,*AR3++(IR0)          ; X(I3) <-----------+
                ADDF3   R6,*AR1,R5              ; R5 = X(I1)+X(I2) ---------+
                LDF     *AR3,R0                 ; R0 = X(I3)                |
                SUBF3   R6,*AR1,R4              ; R4 = X(I1)-X(I2)          |
                SUBF3   R0,R4,R3                ; R3 = X(I1)-X(I2)-X(I3)    |
                ADDF3   R0,R4,R2                ; R2 = X(I1)-X(I2)+X(I3)    |
                SUBF3   R0,*AR4,R1              ; R1 = X(I4)-X(I3) ---------|-+
             || STF     R5,*AR1++(IR0)          ; X(I1) <-------------------+ |
                ADDF3   R2,*AR4,R5              ; R5 = X(I1)-X(I2)+X(I3)+X(I4)|
             || STF     R1,*AR2++(IR0)          ; X(I2) <---------------------+
                MPYF3   R5,*AR7,R1              ; R1 = R5*SIN  <---------------+
             || SUBF3   *AR4,R3,R2              ; R2 = X(I1)-X(I2)-X(I3)-X(I4) |
LOOP3_B:        MPYF3   R2,*AR7,R0              ; R0 = R2*SIN                  |
             || STF     R1,*AR4++(IR0)          ; X(I4) <----------------------+

                STF     R0,*AR3                 ;X(I3)


;
; Perform first and second FFT loops.
;     _         ______
;    |  AR1 -> |__I1__| 0  <-  X(I1) + X(I3) + 2*X(I2)
;    |  AR2 -> |__I2__| 1  <-  X(I1) + X(I3) - 2*X(I2)
;    |  AR3 -> |__I3__| 2  <-  X(I1) - X(I3) - 2*X(I4)
;    |_ AR4 -> |__I4__| 3  <-  X(I1) - X(I3) + 2*X(I4)
;       AR1 -> |______| 4
;              |   .  |
;		   .
;                  .
;                 \|/
;

                LDI     @SOURCE_ADDR,AR1
		LDI	AR1,AR2
		LDI	AR1,AR3
		LDI	AR1,AR4
		ADDI	1,AR2
		ADDI	2,AR3
		ADDI    3,AR4
		LDI	4,IR0
		LDI	@FFT_SIZE,RC
		LSH	-2,RC
		SUBI	2,RC

                LDF     *AR4,R6                 ; R6 = X(I4)
                LDF     *AR2,R7                 ; R7 = X(I2)
             || LDF     *AR1,R1                 ; R1 = X(I1)
                MPYF    2.0,R6                  ; R6 = 2 * X(I4)
                MPYF    2.0,R7                  ; R7 = 2 * X(I2)
                SUBF3   R6,*AR3,R5              ; R5 = X(I3) - 2*X(I4)
                SUBF3   R5,R1,R4                ; R4 = X(I1)-X(I3)+2X(I4) --+
                SUBF3   R7,*AR3,R5              ; R5 = X(I3) - 2*X(I2)      |
             || STF     R4,*AR4++(IR0)          ; X(I4)  <------------------+
                ADDF3   R5,R1,R3                ; R3 = X(I1)+X(I3)-2X(I2) --+
                ADDF3   R6,*AR3,R4              ; R4 = X(I3) + 2*X(I4)      |
             || STF     R3,*AR2++(IR0)          ; X(I2)  <------------------+
                SUBF3   R4,R1,R4                ; R4 = X(I1)-X(I3)-2X(I4) --+
                ADDF3   R7,*AR3,R0              ; R0 = X(I3) + 2*X(I2)      |
             || STF     R4,*AR3++(IR0)          ; X(I3)  <------------------+
                ADDF3   R0,R1,R0                ; R0 = X(I1)+X(I3)+2X(I2) --+
                                                ;                           |
                RPTB    LOOP1_2                 ;                           |
                LDF     *AR4,R6                 ; R6 = X(I4)                |
             || STF     R0,*AR1++(IR0)          ; X(I1)  <------------------+
                MPYF    2.0,R6                  ; R6 = 2 * X(I4)
                LDF     *AR2,R7                 ; R7 = X(I2)
             || LDF     *AR1,R1                 ; R1 = X(I1)
                MPYF    2.0,R7                  ; R7 = 2 * X(I2)
                SUBF3   R6,*AR3,R5              ; R5 = X(I3) - 2*X(I4)
                SUBF3   R5,R1,R4                ; R4 = X(I1)-X(I3)+2X(I4) --+
                SUBF3   R7,*AR3,R5              ; R5 = X(I3) - 2*X(I2)      |
             || STF     R4,*AR4++(IR0)          ; X(I4)  <------------------+
                ADDF3   R5,R1,R3                ; R3 = X(I1)+X(I3)-2X(I2) --+
                ADDF3   R6,*AR3,R4              ; R4 = X(I3) + 2*X(I4)      |
             || STF     R3,*AR2++(IR0)          ; X(I2)  <------------------+
                SUBF3   R4,R1,R4                ; R4 = X(I1)-X(I3)-2X(I4) --+
                ADDF3   R7,*AR3,R0              ; R0 = X(I3) + 2*X(I2)      |
             || STF     R4,*AR3++(IR0)          ; X(I3)  <------------------+
LOOP1_2:        ADDF3   R0,R1,R0                ; R0 = X(I1)+X(I3)+2X(I2) --+
                                                ;                           |
                STF     R0,*AR1                 ; LAST X(I1) <--------------+

;
; Check Bit Reversing Mode (on or off).
;
; BIT_REVERSING =  0, then OFF (no bit reversing).
; BIT_REVERSING <> 0, Then ON.
;
		LDI	@BIT_REVERSE,R0
		CMPI	0,R0
		BZ	MOVE_DATA

;
; Check Bit Reversing Type.
;
; If SourceAddr =  DestAddr, Then In Place Bit Reversing.
; If SourceAddr <> DestAddr, Then Standard Bit Reversing.
;

		LDI	@SOURCE_ADDR,R0
		CMPI	@DEST_ADDR,R0
		BEQ	IN_PLACE

;
; Bit reversing Type 1 (From Source to Destination).
;
; NOTE: abs(SOURCE_ADDR - DEST_ADDR) must be > FFT_SIZE, this is not checked.
;

		LDI	@FFT_SIZE,R0
		SUBI	2,R0
		LDI	@FFT_SIZE,IR0
		LSH	-1,IR0			; IRO = Half FFT size.
		LDI	@SOURCE_ADDR,AR0
		LDI	@DEST_ADDR,AR1

		LDF	*AR0++,R1

		RPTS    R0
 		LDF	*AR0++,R1
	     ||	STF	R1,*AR1++(IR0)B

		STF	R1,*AR1++(IR0)B

                BR      DIVISION

;
; In Place Bit Reversing.
;

		; Bit Reversing On Even Locations, 1st Half Only.

IN_PLACE:	LDI	@FFT_SIZE,IR0
		LSH	-2,IR0		; IRO = Quarter FFT size.
		LDI	2,IR1

		LDI	@FFT_SIZE,RC
		LSH	-2,RC
		SUBI	3,RC
		LDI	@DEST_ADDR,AR0
		LDI	AR0,AR1
		LDI	AR0,AR2

                NOP	*AR1++(IR0)B
		NOP	*AR2++(IR0)B
		LDF	*++AR0(IR1),R0
		LDF	*AR1,R1
		CMPI	AR1,AR0		; Xchange Locations only if AR0<AR1.
		LDFGT	R0,R1
		LDFGT	*AR1++(IR0)B,R1

		RPTB	BITRV1
		LDF 	*++AR0(IR1),R0
	     || STF     R0,*AR0
		LDF     *AR1,R1
	     || STF     R1,*AR2++(IR0)B
		CMPI	AR1,AR0
		LDFGT	R0,R1
BITRV1:		LDFGT   *AR1++(IR0)B,R0

		STF	R0,*AR0
	       	STF	R1,*AR2

		; Perform Bit Reversing On Odd Locations, 2nd Half Only.

		LDI	@FFT_SIZE,RC
		LSH	-1,RC
		LDI	@DEST_ADDR,AR0
		ADDI	RC,AR0
		ADDI	1,AR0
		LDI	AR0,AR1
		LDI	AR0,AR2
		LSH	-1,RC
		SUBI	3,RC

		NOP	*AR1++(IR0)B
		NOP	*AR2++(IR0)B
		LDF	*++AR0(IR1),R0
		LDF	*AR1,R1
		CMPI	AR1,AR0		; Xchange Locations only if AR0<AR1.
		LDFGT	R0,R1
		LDFGT	*AR1++(IR0)B,R1

		RPTB	BITRV2
		LDF 	*++AR0(IR1),R0
	     || STF     R0,*AR0
		LDF     *AR1,R1
	     || STF     R1,*AR2++(IR0)B
		CMPI	AR1,AR0
		LDFGT	R0,R1
BITRV2:		LDFGT   *AR1++(IR0)B,R0

		STF	R0,*AR0
	       	STF	R1,*AR2

		; Perform Bit Reversing On Odd Locations, 1st Half Only.

		LDI	@FFT_SIZE,RC
		LSH	-1,RC
		LDI	RC,IR0
		LDI	@DEST_ADDR,AR0
		LDI	AR0,AR1
		ADDI	1,AR0
		ADDI	IR0,AR1
		LSH	-1,RC
		LDI	RC,IR0
		SUBI	2,RC

		LDF	*AR0,R0
		LDF	*AR1,R1

		RPTB    BITRV3
		LDF	*++AR0(IR1),R0
	     || STF	R0,*AR1++(IR0)B
BITRV3:		LDF	*AR1,R1
	     || STF	R1,*-AR0(IR1)

		STF	R0,*AR1
                STF	R1,*AR0

                BR      DIVISION

;
; Check Data Source Locations.
;
; If SourceAddr =  DestAddr, Then do nothing.
; If SourceAddr <> DestAddr, Then move data.
;

MOVE_DATA:	LDI	@SOURCE_ADDR,R0
		CMPI	@DEST_ADDR,R0
                BEQ     DIVISION

		LDI	@FFT_SIZE,R0
		SUBI	2,R0
		LDI	@SOURCE_ADDR,AR0
		LDI	@DEST_ADDR,AR1

        	LDF	*AR0++,R1

		RPTS    R0
 		LDF	*AR0++,R1
	     ||	STF	R1,*AR1++

		STF	R1,*AR1

DIVISION:       LDI     2,IR0
                LDI     @FFT_SIZE,R0
                FLOAT   R0                  ; exp = LOG_SIZE
                PUSHF   R0                  ; 32 MSB'S saved
                POP     R0
                NEGI    R0                  ; Neg exponent
                PUSH    R0
                POPF    R0                  ; R0 = 1/FFT_SIZE
                LDI     @DEST_ADDR,AR1
                LDI     @DEST_ADDR,AR2
                NOP     *AR2++
                LDI     @FFT_SIZE,RC
                LSH     -1,RC
                SUBI    2,RC
                MPYF3   R0,*AR1,R1          ; 1st location
                RPTB    LAST_LOOP
                MPYF3   R0,*AR2,R2         ; 2nd,4th,6th,... location
             || STF     R1,*AR1++(IR0)
LAST_LOOP:      MPYF3   R0,*AR1,R1         ; 3rd,5th,7th,... location
             || STF     R2,*AR2++(IR0)

                MPYF3   R0,*AR2,R2         ; last location
             || STF     R1,*AR1
                STF     R2,*AR2


; Return to C environment.
;

		POP	DP	; Restore C environment variables.
		POP	AR7
		POP	AR6
		POP	AR5
		POP	AR4
		POPF	R7
		POP	R7
		POPF	R6
		POP	R6
		POP	R5
		POP	R4
		POP	FP
		RETS

		.end

*
* No more.
*
*****************************************************************************
*
