********************************************************************************
*
*  FILENAME    : R2DIT.ASM
*
*  DESCRIPTION : COMPLEX, RADIX-2 DIT FFT FOR TMS320C30
*
*  DATE        : 6/29/93
*
*  VERSION     : 4.0
*
********************************************************************************
*    VERSION        DATE         COMMENTS
*    -------        ----         --------
*     1.0           7/89         Original version
*                                RAIMUND MEYER, KARL SCHWARZ
*                                LEHRSTUHL FUER NACHRICHTENTECHNIK
*                                UNIVERSITAET ERLANGEN-NUERNBERG
*                                CAUERSTRASSE 7, D-8520 ERLANGEN, FRG
*
*    2.0          4/27/92        DANIEL MAZZOCCO   TI HOUSTON
*                                MADE IT C CALLABLE
*                                ALSO CHANGED THE ORDER OF SOURCE ADDRESSES
*                                FOR SOME MPYF3 INSTRUCTIONS IN ORDER TO
*                                HAVE TWO READS IN ONE CYCLE WITH ONE SOURCE
*                                LOCATED EXTERNALLY (TWIDDLE FACTORS )
*
*     3.0          2/25/93       ROSEMARIE PIEDRA (TI HOUSTON): replace
*                                NOP instruction that may cause  dummy external
*                                SRAM access for 1K FFT
*
*     4.0          6/29/93       ROSEMARIE PIEDRA (TI HOUSTON): add support
*                                for in-place bit-reversing
*
********************************************************************************
*									     *
*    SYNOPSIS: int  ffft_r2( FFT_SIZE, LOG_SIZE, SOURCE_ADDR,                *
*                            DEST_ADDR, SINE_TABLE );                        *
*                                                                            *
*            int     FFT_SIZE        ; 64, 128, 256, 512, 1024, ...          *
*            int     LOG_SIZE        ;  6,   7,   8,   9,   10, ...          *
*            float   *SOURCE_ADDR    ; Points to where data is originated    *
*                                    ; and operated on.                      *
*            float   *DEST_ADDR      ; Points to where data will be stored.  *
*            float   *SINE_TABLE     ; Points to the SIN/COS table.          *
*                                                                            *
*    THE (COMPLEX) DATA RESIDE IN INTERNAL MEMORY. THE COMPUTATION IS DONE   *
*    IN-PLACE, BUT THE RESULT IS MOVED TO ANOTHER MEMORY SECTION TO          *
*    DEMONSTRATE THE BIT-REVERSED ADDRESSING.                                *
*    FOR THIS PROGRAM THE MINIMUM FFTLENGTH IS 32 POINTS BECAUSE OF          *
*    THE SEPARATE STAGES.				                     *
*									     *
*    FIRST TWO PASSES ARE REALIZED AS A FOUR BUTTERFLY LOOP SINCE THE        *
*          MULTIPLIES ARE TRIVIAL. THE MULTIPLIER IS ONLY USED FOR           *
*          A LOAD IN PARALLEL WITH AN ADDF OR SUBF.		             *
*    									     *
******************************************************************************
*									     *
*    EXAMPLE FOR A 1024 POINT FFT (EXCLUDING BITREVERSAL):		     *
*									     *
*    MEMORY SIZE :  PROG : 262 WORDS  ;  DATA (TWIDDLE FACTORS): 512 WORDS   *
*     								             *
*    TOTAL NUMBER OF INSTRUCTION CYCLES : 39505                              *
*    TOTAL TIME FOR A 1024 POINT FFT    :  2.37 ms (EXCLUDING BITREVERSAL)   *
*     ( WITH 60NS C3X )                                                      *
******************************************************************************
*									     *
*    THE TWIDDLE FACTORS ARE STORED IN BITREVERSED ORDER AND WITH A TABLE    *
*    LENGTH OF N/2 (N = FFTLENGTH).					     *
*    EXAMPLE: SHOWN FOR N=32, WN(n) = COS(2*PI*n/N) - j*SIN(2*PI*n/N)        *
*									     *
*		ADDRESS			COEFFICIENT			     *
*		   0	      R{WN(0)} = COS(2*PI*0/32) = 1  		     *
*		   1	     -I{WN(0)} = SIN(2*PI*0/32) = 0        	     *
*		   2	      R{WN(4)} = COS(2*PI*4/32) = 0.707		     *
*		   3	     -I{WN(4)} = SIN(2*PI*4/32) = 0.707		     *
*             								     *
*		   :                           :			     *
*             								     *
*		   12	      R{WN(3)} = COS(2*PI*3/32) = 0.831		     *
*		   13	     -I{WN(3)} = SIN(2*PI*3/32) = 0.556		     *
*		   14	      R{WN(7)} = COS(2*PI*7/32) = 0.195		     *
*		   15	     -I{WN(7)} = SIN(2*PI*7/32) = 0.981		     *
*									     *
*    WHEN GENERATED FOR A FFT LENGTH OF 1024, THE TABLE IS FOR ALL FFT       *
*    LENGTH LESS OR EQUAL AVAILABLE.					     *
*    THE MISSING TWIDDLE FACTORS (WN(),WN(),....) ARE GENERATED BY USING     *
*    THE SYMMETRY WN(N/4+n) = -j*WN(n). THIS CAN BE REALIZED VERY EASY, BY   *
*    CHANGING REAL- AND IMAGINARY PART OF THE TWIDDLE FACTORS AND BY         *
*    NEGATING THE NEW REAL PART.  					     *
******************************************************************************
*									     *
*									     *
*                                                +			     *
*   AR + j AI ----------------------------------------------- AR' + j AI'    *
*                                       \         / +			     *
*                                         \     /			     *
*                                           \ /				     *
*                                           / \				     *
*                                         /     \			     *
*                                       /         \ +			     *
*   BR + j BI ---- ( COS - j SIN ) -------------------------- BR' + j BI'    *
*                                                -			     *
*									     *
*   TR = BR * COS + BI * SIN						     *
*   TI = BR * SIN - BI * COS  						     *
*   AR'= AR + TR							     *
*   AI'= AI - TI							     *
*   BR'= AR - TR							     *
*   BI'= AI + TI							     *
*									     *
******************************************************************************
FP		.set	AR3
                .global _ffft_r2         ; Entry execution point.
                .global _halfsize,_quartsize_minus2,_octsize_minus2
                                         ; Defined in the main C file

                .bss   FFT_SIZE    ,1    ; Reserve memory for arguments.
                .bss   LOG_SIZE    ,1
                .bss   SOURCE_ADDR ,1
                .bss   DEST_ADDR   ,1
                .bss   SINE_TABLE  ,1
                .bss   sintm1  ,1
                .bss   sintp2  ,1
                .bss   inputp2 ,1
                .bss   outp1   ,1

                .sect   ".cfftdat"
fg2             .set    _halfsize         ; is FFT_SIZE/2
fg4m2           .set    _quartsize_minus2 ; is FFT_SIZE/4 - 2
fg8m2           .set    _octsize_minus2   ; is FFT_SIZE/8 - 2
;
; Initialise C Function.
;
                .sect   ".cffttxt"

_ffft_r2:       PUSH    FP              ; Preserve C environment.
		LDI	SP,FP
		PUSH	R4
		PUSH	R5
                PUSH	R6
		PUSHF	R6
		PUSH	R7
		PUSHF	R7
		PUSH	AR4
		PUSH	AR5
		PUSH	AR6
		PUSH	AR7
		PUSH	DP

                LDP     FFT_SIZE        ; Initialise DP pointer.
		LDI	*-FP(2),R0	; Move arguments from stack.
                STI     R0,@FFT_SIZE
		LDI	*-FP(3),R0
		STI	R0,@LOG_SIZE
		LDI	*-FP(4),R0
		STI	R0,@SOURCE_ADDR
                ADDI    2,R0
                STI     R0,@inputp2     ; inputp2 is SOURCE_ADDR plus 2.
		LDI	*-FP(5),R0
		STI	R0,@DEST_ADDR
		LDI	*-FP(6),R0
		STI	R0,@SINE_TABLE
                SUBI    1,R0
                STI     R0,@sintm1      ; sintm1 is SINE_TABLE minus 1
                ADDI    3,R0
                STI     R0,@sintp2      ; sintp2 is SINE_TABLE plus 2





*       ar0 : AR + AI
*       ar1 : BR + BI
*       ar2 : CR + CI + CR' + CI'
*       ar3 : DR + DI
*       ar4 : AR' + AI'
*       ar5 : BR' + BI'
*       ar6 : DR' + DI'
*       ar7 : first twiddle factor = 1

        ldi     @fg2,ir0                ; ir0 = n/2 = offset between SOURCE_ADDRs
        ldi     @SINE_TABLE,ar7         ; ar7 points to twiddle factor 1
        ldi     @SOURCE_ADDR,ar0        ; ar0 points to AR
        addi    ir0,ar0,ar1             ; ar1 points to BR
        addi    ir0,ar1,ar2             ; ar2 points to CR
        addi    ir0,ar2,ar3             ; ar3 points to DR
        ldi     ar0,ar4                 ; ar4 points to AR'
        ldi     ar1,ar5                 ; ar5 points to BR'
        ldi     ar3,ar6                 ; ar6 points to DR'
	ldi	2,ir1			; addressoffset
        lsh     -1,ir0                  ; ir0 = n/4 = number of R4-butterflies
        ldi     ir0,rc
        subi    2,rc

*****************************************************************************
* ------------ FIRST 2 STAGES AS RADIX-4 BUTTERFLY -------------------------*
*****************************************************************************

* fill pipeline

        addf    *ar2,*ar0,r4            ; r4 = AR + CR       
        subf    *ar2,*ar0++,r5          ; r5 = AR - CR
        addf    *ar1,*ar3,r6            ; r6 = DR + BR
        subf    *ar1++,*ar3++,r7        ; r7 = DR - BR
        addf    r6,r4,r0                ; AR' = r0 = r4 + r6
        mpyf    *ar7,*ar3++,r1          ; r1 = DI , BR' = r3 = r4 - r6
||      subf    r6,r4,r3
        addf    r1,*ar1,r0              ; r0 = BI + DI , AR' = r0
||      stf     r0,*ar4++
        subf    r1,*ar1++,r1            ; r1 = BI - DI , BR' = r3
||      stf     r3,*ar5++
        addf    r1,r5,r2                ; CR' = r2 = r5 + r1
        mpyf    *ar7,*+ar2,r1           ; r1 = CI , DR' = r3 = r5 - r1
||      subf    r1,r5,r3
        addf    r1,*ar0,r2              ; r2 = AI + CI , CR' = r2
||      stf     r2,*ar2++(ir1)
        subf    r1,*ar0++,r6            ; r6 = AI - CI , DR' = r3      
||      stf     r3,*ar6++ 
        addf    r0,r2,r4                ; AI' = r4 = r2 + r0

* radix-4 butterfly loop

        rptb    blk1
        mpyf    *ar7,*ar2--,r0          ; r0 = CR , (BI' = r2 = r2 - r0)
||      subf    r0,r2,r2
        mpyf    *ar7,*ar1++,r1          ; r1 = BR , (CI' = r3 = r6 + r7)
||      addf    r7,r6,r3
        addf    r0,*ar0,r4              ; r4 = AR + CR , (AI' = r4)
||      stf     r4,*ar4++
        subf    r0,*ar0++,r5            ; r5 = AR - CR , (BI' = r2)
||      stf     r2,*ar5++
        subf    r7,r6,r7                ; (DI' = r7 = r6 - r7)
        addf    r1,*ar3,r6              ; r6 = DR + BR , (DI' = r7)
||      stf     r7,*ar6++
        subf    r1,*ar3++,r7            ; r7 = DR - BR , (CI' = r3)
||      stf     r3,*ar2++
        addf    r6,r4,r0                ; AR' = r0 = r4 + r6
        mpyf    *ar7,*ar3++,r1          ; r1 = DI , BR' = r3 = r4 - r6
||      subf    r6,r4,r3
        addf    r1,*ar1,r0              ; r0 = BI + DI , AR' = r0
||      stf     r0,*ar4++
        subf    r1,*ar1++,r1            ; r1 = BI - DI , BR' = r3
||      stf     r3,*ar5++
        addf    r1,r5,r2                ; CR' = r2 = r5 + r1
        mpyf    *ar7,*+ar2,r1           ; r1 = CI , DR' = r3 = r5 - r1
||      subf    r1,r5,r3
        addf    r1,*ar0,r2              ; r2 = AI + CI , CR' = r2
||      stf     r2,*ar2++(ir1)
        subf    r1,*ar0++,r6            ; r6 = AI - CI , DR' = r3      
||      stf     r3,*ar6++
blk1    addf    r0,r2,r4                ; AI' = r4 = r2 + r0

* clear pipeline

        subf    r0,r2,r2                ; BI' = r2 = r2 - r0
        addf    r7,r6,r3                ; CI' = r3 = r6 + r7
        stf     r4,*ar4                 ; AI' = r4 , BI' = r2
||      stf     r2,*ar5
        subf    r7,r6,r7                ; DI' = r7 = r6 - r7
        stf     r7,*ar6                 ; DI' = r7 , CI' = r3
||      stf     r3,*--ar2

*****************************************************************************
* ------------ THIRD TO LAST-2 STAGE ---------------------------------------*
*****************************************************************************

	ldi	@fg2,ir1
	ldi	ir0,ar5
	subi	1,ar5
	ldi	1,ar6

stufe   ldi     @SINE_TABLE,ar7 ; pointer to twiddle factor
	ldi	0,ar4		; group counter
        ldi     @SOURCE_ADDR,ar0      ; upper real butterfly SOURCE_ADDR
	ldi	ar0,ar2		; upper real butterfly output
	addi	ir0,ar0,ar3	; lower real butterfly output
        ldi     ar3,ar1         ; lower real butterfly SOURCE_ADDR
	lsh	1,ar6		; double group count 
	lsh	-2,ar5		; half butterfly count
	lsh	1,ar5		; clear LSB
	lsh	-1,ir0		; half step from upper to lower real part
	lsh	-1,ir1
	addi	1,ir1		; step from old imaginary to new real value
	ldf	*ar1++,r6	; dummy load, only for address update
||	ldf	*ar7,r7		; r7 = COS

gruppe

* fill pipeline                 ; ar0 = upper real butterfly SOURCE_ADDR
                                ; ar1 = lower real butterfly SOURCE_ADDR
				; ar2 = upper real butterfly output
				; ar3 = lower real butterfly output
				; the imaginary part has to follow
	ldf	*++ar7,r6		; r6 = SIN
	mpyf	*ar1--,r6,r1		; r1 = BI * SIN
||	addf	*++ar4,r0,r3		; dummy addf for counter update
	mpyf	*ar1,r7,r0		; r0 = BR * COS
        mpyf    *ar7--,*ar1++,r0        ; r3 = TR = r0 + r1 , r0 = BR * SIN
||	addf	r0,r1,r3	
	mpyf	*ar1++,r7,r1		; r1 = BI * COS , r2 = AR - TR
||	subf	r3,*ar0,r2		
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++		 
	ldi	ar5,rc

*   FIRST BUTTERFLY-TYPE:
*									     *
*   TR = BR * COS + BI * SIN						     *
*   TI = BR * SIN - BI * COS  						     *
*   AR'= AR + TR							     *
*   AI'= AI - TI							     *
*   BR'= AR - TR							     *
*   BI'= AI + TI							     *

	rptb	bfly1

	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r5)
||	stf	r5,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1) 
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r7,r1		; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++
bfly1	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

* switch over to next group

	subf	r1,r0,r2		; r2 = TI = r0 - r1
	addf	r2,*ar0,r3		; r3 = AI + TI , AR' = r5
||	stf	r5,*ar2++	
	subf	r2,*ar0++(ir1),r4	; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3++(ir1)		
	nop	*ar1++(ir1)		; address update
	mpyf	*ar1--,r7,r1		; r1 = BI * COS , AI' = r4
||	stf	r4,*ar2++(ir1)		
	mpyf	*ar1,r6,r0		; r0 = BR * SIN
        mpyf    *ar7++,*ar1++,r0        ; r3 = TR = r1 - r0 , r0 = BR * COS
||	subf	r0,r1,r3	
	mpyf	*ar1++,r6,r1		; r1 = BI * SIN , r2 = AR - TR
||	subf	r3,*ar0,r2		
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++		 
	ldi	ar5,rc

*   SECOND BUTTERFLY-TYPE:
*									     *
*   TR = BI * COS - BR * SIN						     *
*   TI = BI * SIN + BR * COS  						     *
*   AR'= AR + TR							     *
*   AI'= AI - TI							     *
*   BR'= AR - TR							     *
*   BI'= AI + TI							     *

	rptb	bfly2

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r5)
||	stf	r5,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1) 
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++
	subf	r0,r5,r3		; TR = r3 = r5 - r0
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r6,r1		; r1 = BI * SIN , (AI' = r4)
||	stf	r4,*ar2++
bfly2	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

* clear pipeline

	addf	r1,r0,r2		; r2 = TI = r0 + r1
	addf	r2,*ar0,r3		; r3 = AI + TI
||	stf	r5,*ar2++		; AR' = r5
	cmpi	ar6,ar4
	bned	gruppe			; do following 3 instructions
	subf	r2,*ar0++(ir1),r4	; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3++(ir1)	
	ldf	*++ar7,r7		; r7 = COS
||	stf	r4,*ar2++(ir1)		; AI' = r4
        addi    ar1,ir1,ar1             ; version 3.0 fix : replace instruction
                                        ; nop     *ar1++(ir1)

* end of this butterflygroup

	cmpi	4,ir0			; jump out after ld(n)-3 stage
	bnz	stufe

*****************************************************************************
*------------- SECOND LAST STAGE -------------------------------------------*
*****************************************************************************

        ldi     @SOURCE_ADDR,ar0        ; upper SOURCE_ADDR
	ldi	ar0,ar2			; upper output
        addi    ir0,ar0,ar1             ; lower SOURCE_ADDR
	ldi	ar1,ar3			; lower output
	ldi	@sintp2,ar7		; pointer to twiddle faktor
	ldi	5,ir0			; distance between two groups
	ldi	@fg8m2,rc

* fill pipeline 

* 1. butterfly: w^0
	addf	*ar0,*ar1,r2		; AR' = r2 = AR + BR
	subf	*ar1++,*ar0++,r3	; BR' = r3 = AR - BR
	addf	*ar0,*ar1,r0		; AI' = r0 = AI + BI
	subf	*ar1++,*ar0++,r1	; BI' = r1 = AI - BI
* 2. butterfly: w^0
	addf	*ar0,*ar1,r6		; AR' = r6 = AR + BR
	subf	*ar1++,*ar0++,r7	; BR' = r7 = AR - BR
	addf	*ar0,*ar1,r4		; AI' = r4 = AI + BI
	subf	*ar1++(ir0),*ar0++(ir0),r5	; BI' = r5 = AI - BI
	stf	r2,*ar2++		; (AR' = r2)
||	stf	r3,*ar3++		; (BR' = r3)
	stf	r0,*ar2++		; (AI' = r0)
||	stf	r1,*ar3++		; (BI' = r1)
	stf	r6,*ar2++		; AR' = r6
||	stf	r7,*ar3++		; BR' = r7
	stf	r4,*ar2++(ir0)		; AI' = r4
||	stf	r5,*ar3++(ir0)		; BI' = r5
* 3. butterfly: w^M/4
	addf	*ar0++,*+ar1,r5		; AR' = r5 = AR + BI
	subf	*ar1,*ar0,r4		; AI' = r4 = AI - BR
	addf	*ar1++,*ar0--,r6	; BI' = r6 = AI + BR
	subf	*ar1++,*ar0++,r7	; BR' = r7 = AR - BI
* 4. butterfly: w^M/4
	addf	*+ar1,*++ar0,r3		; AR' = r3 = AR + BI 
	ldf	*-ar7,r1		; r1 = 0 (for inner loop)
||	ldf	*ar1++,r0		; r0 = BR (for inner loop)
	subf	*ar1++(ir0),*ar0++,r2	; BR' = r2 = AR - BI 
	stf	r5,*ar2++		; (AR' = r5)
||	stf	r7,*ar3++		; (BR' = r7)
	stf	r6,*ar3++		; (BI' = r6)

* 5. to M. butterfly: 

	rptb	bf2end	

	ldf	*ar7++,r7		; r7 = COS , ((AI' = r4))
||	stf	r4,*ar2++
	ldf	*ar7++,r6		; r6 = SIN , (BR' = r2)
||	stf	r2,*ar3++
	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r3)
||	stf	r3,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1) 
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++(ir0),r4	; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++(ir0)
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r7,r1		; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++(ir0)
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r5)
||	stf	r5,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1) 
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++(ir0),r7,r1	; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++
	addf	*ar0++,r3,r3	 	; r3 = AR + TR , BR' = r2
||	stf	r2,*ar3++

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r3)
||	stf	r3,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1) 
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++(ir0),r4	; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++(ir0)
	subf	r0,r5,r3		; r3 = TR = r5 - r0
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r6,r1		; r1 = BI * SIN , (AI' = r4)
||	stf	r4,*ar2++(ir0)
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r5)
||	stf	r5,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1) 
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , y(L) = BI' = r3)
||	stf	r3,*ar3++
	subf	r0,r5,r3		; r3 = TR = r5 - r0
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2
bf2end	mpyf	*ar1++(ir0),r6,r1	; r1 = BI * SIN , r3 = AR + TR
||	addf	*ar0++,r3,r3	

* clear pipeline

	stf	r2,*ar3++		; BR' = r2 , AI' = r4
||	stf	r4,*ar2++
	addf	r1,r0,r2		; r2 = TI = r0 + r1
	addf	r2,*ar0,r3		; r3 = AI + TI , AR' = r3
||	stf	r3,*ar2++		
	subf	r2,*ar0,r4		; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3	
	stf	r4,*ar2			; AI' = r4

*****************************************************************************
*------------- LAST STAGE --------------------------------------------------*
*****************************************************************************

        ldi     @SOURCE_ADDR,ar0   ; upper SOURCE_ADDR
        ldi     ar0,ar2            ; upper output
        ldi     @inputp2,ar1 ; lower SOURCE_ADDR
        ldi     ar1,ar3            ; lower output
        ldi     @sintp2,ar7        ; pointer to twiddle factors
        ldi     3,ir0              ; group offset
	ldi	@fg4m2,rc

* fill pipeline 

* 1. butterfly: w^0
	addf	*ar0,*ar1,r6		; AR' = r6 = AR + BR
	subf	*ar1++,*ar0++,r7	; BR' = r7 = AR - BR
	addf	*ar0,*ar1,r4		; AI' = r4 = AI + BI
	subf	*ar1++(ir0),*ar0++(ir0),r5	; BI' = r5 = AI - BI
* 2. butterfly: w^M/4
	addf	*+ar1,*ar0,r3		; AR' = r3 = AR + BI
	ldf	*-ar7,r1		; r1 = 0 (for inner loop) 
||	ldf	*ar1++,r0		; r0 = BR (for inner loop)
	subf	*ar1++(ir0),*ar0++,r2	; BR' = r2 = AR - BI 
	stf	r6,*ar2++		; (AR' = r6)
||	stf	r7,*ar3++		; (BR' = r7)
	stf	r5,*ar3++(ir0)		; (BI' = r5)

* 3. to M. butterfly: 
	rptb	bflend	

	ldf	*ar7++,r7		; r7 = COS , ((AI' = r4))
||	stf	r4,*ar2++(ir0)
	ldf	*ar7++,r6		; r6 = SIN , (BR' = r2)
||	stf	r2,*ar3++
	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r3)  
||	stf	r3,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1)
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++(ir0),r4	; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++(ir0)
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR
||	subf	r3,*ar0,r2 
	mpyf	*ar1++(ir0),r7,r1	; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++(ir0)
	addf	*ar0++,r3,r3		; r3 = AR + TR , BR' = r2
||	stf	r2,*ar3++

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r3)  
||	stf	r3,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1)
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++(ir0),r4	; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++(ir0)
	subf	r0,r5,r3		; r3 = TR = r0 - r5
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2 
bflend	mpyf	*ar1++(ir0),r6,r1	; r1 = BI * SIN , r3 = AR + TR
||	addf	*ar0++,r3,r3	

* clear pipeline

	stf	r2,*ar3++		; BR' = r2 , (AI' = r4)
||	stf	r4,*ar2++(ir0)
	addf	r1,r0,r2		; r2 = TI = r0 + r1
	addf	r2,*ar0,r3		; r3 = AI + TI , AR' = r3
||	stf	r3,*ar2++
	subf	r2,*ar0,r4		; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3			
	stf	r4,*ar2			; AI' = r4

*****************************************************************************
*------------- END OF FFT --------------------------------------------------*
*****************************************************************************

ENDB:
*****************************************************************************
*------------- BITREVERSAL -------------------------------------------------*
*****************************************************************************
        ldi     @SOURCE_ADDR,ar0
        cmpi    @DEST_ADDR,ar0
        beqd    inplace
        ldi     @DEST_ADDR,ar1           ; ar1: DST_ADDR
        ldi     @FFT_SIZE,ir0           ; ir0: FFT_SIZE
        ldi     ir0,rc

	ldi	2,ir1
	subi	2,rc

	ldf	*+ar0(1),r0
        rptb    bitrv1
	ldf	*ar0++(ir0)b,r1
||	stf	r0,*+ar1(1)
bitrv1  ldf     *+ar0(1),r0
||	stf	r1,*ar1++(ir1)

        bud     end
	ldf	*ar0++(ir0)b,r1
||	stf	r0,*+ar1(1)
	stf	r1,*ar1

inplace

	subi	2,rc
        nop     *++ar1(2)
        nop     *ar0++(ir0)b
        rptb    bitrv2                  ;

        cmpi    ar1,ar0
        bge     cont
        ldf     *ar1,r0
||      ldf     *ar0,r1
        stf     r0,*ar0
||      stf     r1,*ar1
        ldf     *+ar1(1),r0
||      ldf     *+ar0(1),r1

        stf     r0,*+ar0(1)
||      stf     r1,*+ar1(1)
cont    nop     *++ar1(2)
bitrv2  nop     *ar0++(ir0)b


;
; Return to C environment.
;

*****************************************************************************
*------------- END OF FFT --------------------------------------------------*
*****************************************************************************

end:

		POP	DP	; Restore C environment variables.
		POP	AR7
		POP	AR6
		POP	AR5
		POP	AR4
		POPF	R7
		POP	R7
		POPF	R6
		POP	R6
		POP	R5
		POP	R4
		POP	FP
		RETS
	.end


