******************************************************************************
*									     *
*                COMPLEX, RADIX-2 DIT FFT  :  R2DITB.ASM   	             *
*             --------------------------------------------     		     *
*								             *
*    GENERIC PROGRAM FOR A FAST LOOPED-CODE RADIX-2 DIT FFT COMPUTATION      *
*                          IN TMS320C30				             *
*									     *
*    WRITTEN BY: RAIMUND MEYER, KARL SCHWARZ               24.07.89	     *
*									     *
*		 LEHRSTUHL FUER NACHRICHTENTECHNIK			     *
*		 UNIVERSITAET ERLANGEN-NUERNBERG			     *
*                CAUERSTRASSE 7, D-8520 ERLANGEN, FRG                        *
*									     *
*    REVISION:  							     *
*									     *
*    THE (COMPLEX) DATA RESIDE IN INTERNAL MEMORY. THE COMPUTATION IS DONE   *
*    IN-PLACE, BUT THE RESULT IS MOVED TO ANOTHER MEMORY SECTION TO          *
*    DEMONSTRATE THE BIT-REVERSED ADDRESSING.                                *
*    FOR THIS PROGRAM THE MINIMUM FFTLENGTH IS 32 POINTS BECAUSE OF          *
*    THE SEPARATE STAGES.				                     *
*									     *
*    FIRST TWO PASSES ARE REALIZED AS A FOUR BUTTERFLY LOOP SINCE THE        *
*          MULTIPLIES ARE TRIVIAL. THE MULTIPLIER IS ONLY USED FOR           *
*          A LOAD IN PARALLEL WITH AN ADDF OR SUBF.		             *
*    									     *
******************************************************************************
*									     *
*    EXAMPLE FOR A 1024 POINT FFT (WITH BITREVERSAL) :			     *
*									     *
*    MEMORY SIZE :  PROG : 231 WORDS  ;  DATA : 512 WORDS  	             *
*     								             *
*    CYCLES PER BUTTERFLY :						     *
*    1. AND 2. STAGE			:     4				     *
*    3. TO  8. STAGE			:     8				     *
*    9.  STAGE				:     8.25			     *
*    10. STAGE				:    10.5 (DUE TO EXT. MEMORY WAITS) *
*    AVERAGE CYCLES/BUTTERFLY           :     7.475 			     *
*    TOTAL BUTTERFLYCYCLES              : 38272				     *
*    INITIALIZATION OVERHEAD		:  2185  =  5.4 % OF TOTAL TIME      *
*    TOTAL NUMBER OF INSTRUCTION CYCLES : 40457				     *
*    TOTAL TIME FOR A 1024 POINT FFT    :  2.42 ms (INCLUDING BITREVERSAL)   *
*									     *
******************************************************************************
*									     *
*    THIS PROGRAM INCUDES FOLLOWING FILES:				     *
*    -------------------------------------				     *
*    THE FILE 'TWID1KBR.ASM' CONSISTS OF TWIDDLE FACTORS		     *
*									     *
*    THE TWIDDLE FACTORS ARE STORED IN BITREVERSED ORDER AND WITH A TABLE    *
*    LENGTH OF N/2 (N = FFTLENGTH).					     *
*    EXAMPLE: SHOWN FOR N=32, WN(n) = COS(2*PI*n/N) - j*SIN(2*PI*n/N)        *
*									     *
*		ADDRESS			COEFFICIENT			     *
*		   0	      R{WN(0)} = COS(2*PI*0/32) = 1  		     *
*		   1	     -I{WN(0)} = SIN(2*PI*0/32) = 0        	     *
*		   2	      R{WN(4)} = COS(2*PI*4/32) = 0.707		     *
*		   3	     -I{WN(4)} = SIN(2*PI*4/32) = 0.707		     *
*             								     *
*		   :                           :			     *
*             								     *
*		   12	      R{WN(3)} = COS(2*PI*3/32) = 0.831		     *
*		   13	     -I{WN(3)} = SIN(2*PI*3/32) = 0.556		     *
*		   14	      R{WN(7)} = COS(2*PI*7/32) = 0.195		     *
*		   15	     -I{WN(7)} = SIN(2*PI*7/32) = 0.981		     *
*									     *
*    WHEN GENERATED FOR A FFT LENGTH OF 1024, THE TABLE IS FOR ALL FFT       *
*    LENGTH LESS OR EQUAL AVAILABLE.					     *
*    THE MISSING TWIDDLE FACTORS (WN(),WN(),....) ARE GENERATED BY USING     *
*    THE SYMMETRY WN(N/4+n) = -j*WN(n). THIS CAN BE REALIZED VERY EASY, BY   *
*    CHANGING REAL- AND IMAGINARY PART OF THE TWIDDLE FACTORS AND BY         *
*    NEGATING THE NEW REAL PART.  					     *
*    TO CHANGE THE FFTLENGTH ONLY THE PARAMETERS IN THE HEADER OF 	     *
*    TWID1KBR.ASM AND THE INPUT AND OUTPUT VECTOR LENGTH NEED TO BE ALTERED. *
******************************************************************************
*									     *
*									     *
*                                                +			     *
*   AR + j AI ----------------------------------------------- AR' + j AI'    *
*                                       \         / +			     *
*                                         \     /			     *
*                                           \ /				     *
*                                           / \				     *
*                                         /     \			     *
*                                       /         \ +			     *
*   BR + j BI ---- ( COS - j SIN ) -------------------------- BR' + j BI'    *
*                                                -			     *
*									     *
*   TR = BR * COS + BI * SIN						     *
*   TI = BR * SIN - BI * COS  						     *
*   AR'= AR + TR							     *
*   AI'= AI - TI							     *
*   BR'= AR - TR							     *
*   BI'= AI + TI							     *
*									     *
******************************************************************************

	.global fft
	.global n
	.global nhalb
	.global nviert
	.global nachtel
	.global m
	.global sine

	.BSS	inp,2048		; input  vector length = 2n (depends of n)
	.BSS 	outp,2048		; output vector length = 2n (depends of n)

	.text

fftsiz	.word	n
fg4m2	.word	nviert-2
fg4m3	.word	nviert-3
fg8m2	.word	nachtel-2
fg2	.word	nhalb
fg2m3	.word	nhalb-3
logfft	.word	m	
sintab	.word	sine
sintm1	.word	sine-1
sintp2	.word	sine+2
input	.word	inp	
inputp2	.word	inp+2
output	.word	outp	
outp1	.word	outp+1

*       ar0 : AR + AI
*       ar1 : BR + BI
*       ar2 : CR + CI + CR' + CI'
*       ar3 : DR + DI
*       ar4 : AR' + AI'
*       ar5 : BR' + BI'
*       ar6 : DR' + DI'
*       ar7 : first twiddle factor = 1

fft:    ldp     fftsiz                  ; load page pointer
        ldi     @fg2,ir0                ; ir0 = n/2 = offset between inputs
        ldi     @sintab,ar7             ; ar7 points to twiddle factor 1
        ldi     @input,ar0              ; ar0 points to AR
        addi    ir0,ar0,ar1             ; ar1 points to BR
        addi    ir0,ar1,ar2             ; ar2 points to CR
        addi    ir0,ar2,ar3             ; ar3 points to DR
        ldi     ar0,ar4                 ; ar4 points to AR'
        ldi     ar1,ar5                 ; ar5 points to BR'
        ldi     ar3,ar6                 ; ar6 points to DR'
	ldi	2,ir1			; addressoffset
        lsh     -1,ir0                  ; ir0 = n/4 = number of R4-butterflies
        ldi     ir0,rc
        subi    2,rc

*****************************************************************************
* ------------ FIRST 2 STAGES AS RADIX-4 BUTTERFLY ------------------------ *
*****************************************************************************

* fill pipeline

        addf    *ar2,*ar0,r4            ; r4 = AR + CR       
        subf    *ar2,*ar0++,r5          ; r5 = AR - CR
        addf    *ar1,*ar3,r6            ; r6 = DR + BR
        subf    *ar1++,*ar3++,r7        ; r7 = DR - BR
        addf    r6,r4,r0                ; AR' = r0 = r4 + r6
        mpyf    *ar3++,*ar7,r1          ; r1 = DI , BR' = r3 = r4 - r6
||      subf    r6,r4,r3
        addf    r1,*ar1,r0              ; r0 = BI + DI , AR' = r0
||      stf     r0,*ar4++
        subf    r1,*ar1++,r1            ; r1 = BI - DI , BR' = r3
||      stf     r3,*ar5++
        addf    r1,r5,r2                ; CR' = r2 = r5 + r1
        mpyf    *+ar2,*ar7,r1           ; r1 = CI , DR' = r3 = r5 - r1
||      subf    r1,r5,r3
        addf    r1,*ar0,r2              ; r2 = AI + CI , CR' = r2
||      stf     r2,*ar2++(ir1)
        subf    r1,*ar0++,r6            ; r6 = AI - CI , DR' = r3      
||      stf     r3,*ar6++ 
        addf    r0,r2,r4                ; AI' = r4 = r2 + r0

* radix-4 butterfly loop

        rptb    blk1
        mpyf    *ar2--,*ar7,r0          ; r0 = CR , (BI' = r2 = r2 - r0)
||      subf    r0,r2,r2
        mpyf    *ar1++,*ar7,r1          ; r1 = BR , (CI' = r3 = r6 + r7)
||      addf    r7,r6,r3
        addf    r0,*ar0,r4              ; r4 = AR + CR , (AI' = r4)
||      stf     r4,*ar4++
        subf    r0,*ar0++,r5            ; r5 = AR - CR , (BI' = r2)
||      stf     r2,*ar5++
        subf    r7,r6,r7                ; (DI' = r7 = r6 - r7)
        addf    r1,*ar3,r6              ; r6 = DR + BR , (DI' = r7)
||      stf     r7,*ar6++
        subf    r1,*ar3++,r7            ; r7 = DR - BR , (CI' = r3)
||      stf     r3,*ar2++
        addf    r6,r4,r0                ; AR' = r0 = r4 + r6
        mpyf    *ar3++,*ar7,r1          ; r1 = DI , BR' = r3 = r4 - r6
||      subf    r6,r4,r3
        addf    r1,*ar1,r0              ; r0 = BI + DI , AR' = r0
||      stf     r0,*ar4++
        subf    r1,*ar1++,r1            ; r1 = BI - DI , BR' = r3
||      stf     r3,*ar5++
        addf    r1,r5,r2                ; CR' = r2 = r5 + r1
        mpyf    *+ar2,*ar7,r1           ; r1 = CI , DR' = r3 = r5 - r1
||      subf    r1,r5,r3
        addf    r1,*ar0,r2              ; r2 = AI + CI , CR' = r2
||      stf     r2,*ar2++(ir1)
        subf    r1,*ar0++,r6            ; r6 = AI - CI , DR' = r3      
||      stf     r3,*ar6++
blk1    addf    r0,r2,r4                ; AI' = r4 = r2 + r0

* clear pipeline

        subf    r0,r2,r2                ; BI' = r2 = r2 - r0
        addf    r7,r6,r3                ; CI' = r3 = r6 + r7
        stf     r4,*ar4                 ; AI' = r4 , BI' = r2
||      stf     r2,*ar5
        subf    r7,r6,r7                ; DI' = r7 = r6 - r7
        stf     r7,*ar6                 ; DI' = r7 , CI' = r3
||      stf     r3,*--ar2

*****************************************************************************
* ------------ THIRD TO LAST-2 STAGE -------------------------------------- *
*****************************************************************************

	ldi	@fg2,ir1
	ldi	ir0,ar5
	subi	1,ar5
	ldi	1,ar6

stufe	ldi	@sintab,ar7	; pointer to twiddle factor
	ldi	0,ar4		; group counter
	ldi	@input,ar0	; upper real butterfly input
	ldi	ar0,ar2		; upper real butterfly output
	addi	ir0,ar0,ar3	; lower real butterfly output
	ldi	ar3,ar1		; lower real butterfly input
	lsh	1,ar6		; double group count 
	lsh	-2,ar5		; half butterfly count
	lsh	1,ar5		; clear LSB
	lsh	-1,ir0		; half step from upper to lower real part
	lsh	-1,ir1
	addi	1,ir1		; step from old imaginary to new real value
	ldf	*ar1++,r6	; dummy load, only for address update
||	ldf	*ar7,r7		; r7 = COS

gruppe

* fill pipeline 		; ar0 = upper real butterfly input
				; ar1 = lower real butterfly input
				; ar2 = upper real butterfly output
				; ar3 = lower real butterfly output
				; the imaginary part has to follow
	ldf	*++ar7,r6		; r6 = SIN
	mpyf	*ar1--,r6,r1		; r1 = BI * SIN
||	addf	*++ar4,r0,r3		; dummy addf for counter update
	mpyf	*ar1,r7,r0		; r0 = BR * COS
	mpyf	*ar1++,*ar7--,r0	; r3 = TR = r0 + r1 , r0 = BR * SIN
||	addf	r0,r1,r3	
	mpyf	*ar1++,r7,r1		; r1 = BI * COS , r2 = AR - TR
||	subf	r3,*ar0,r2		
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++		 
	ldi	ar5,rc

*   FIRST BUTTERFLY-TYPE:
*									     *
*   TR = BR * COS + BI * SIN						     *
*   TI = BR * SIN - BI * COS  						     *
*   AR'= AR + TR							     *
*   AI'= AI - TI							     *
*   BR'= AR - TR							     *
*   BI'= AI + TI							     *

	rptb	bfly1

	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r5)
||	stf	r5,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1) 
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r7,r1		; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++
bfly1	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

* switch over to next group

	subf	r1,r0,r2		; r2 = TI = r0 - r1
	addf	r2,*ar0,r3		; r3 = AI + TI , AR' = r5
||	stf	r5,*ar2++	
	subf	r2,*ar0++(ir1),r4	; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3++(ir1)		
	nop	*ar1++(ir1)		; address update
	mpyf	*ar1--,r7,r1		; r1 = BI * COS , AI' = r4
||	stf	r4,*ar2++(ir1)		
	mpyf	*ar1,r6,r0		; r0 = BR * SIN
	mpyf	*ar1++,*ar7++,r0	; r3 = TR = r1 - r0 , r0 = BR * COS
||	subf	r0,r1,r3	
	mpyf	*ar1++,r6,r1		; r1 = BI * SIN , r2 = AR - TR
||	subf	r3,*ar0,r2		
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++		 
	ldi	ar5,rc

*   SECOND BUTTERFLY-TYPE:
*									     *
*   TR = BI * COS - BR * SIN						     *
*   TI = BI * SIN + BR * COS  						     *
*   AR'= AR + TR							     *
*   AI'= AI - TI							     *
*   BR'= AR - TR							     *
*   BI'= AI + TI							     *

	rptb	bfly2

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r5)
||	stf	r5,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1) 
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++
	subf	r0,r5,r3		; TR = r3 = r5 - r0
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r6,r1		; r1 = BI * SIN , (AI' = r4)
||	stf	r4,*ar2++
bfly2	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

* clear pipeline

	addf	r1,r0,r2		; r2 = TI = r0 + r1
	addf	r2,*ar0,r3		; r3 = AI + TI
||	stf	r5,*ar2++		; AR' = r5
	cmpi	ar6,ar4
	bned	gruppe			; do following 3 instructions
	subf	r2,*ar0++(ir1),r4	; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3++(ir1)	
	ldf	*++ar7,r7		; r7 = COS
||	stf	r4,*ar2++(ir1)		; AI' = r4
	nop	*ar1++(ir1)		; branch here

* end of this butterflygroup

	cmpi	4,ir0			; jump out after ld(n)-3 stage
	bnz	stufe

*****************************************************************************
* ------------ SECOND LAST STAGE ------------------------------------------ *
*****************************************************************************

	ldi	@input,ar0		; upper input
	ldi	ar0,ar2			; upper output
	addi	ir0,ar0,ar1		; lower input
	ldi	ar1,ar3			; lower output
	ldi	@sintp2,ar7		; pointer to twiddle faktor
	ldi	5,ir0			; distance between two groups
	ldi	@fg8m2,rc

* fill pipeline 

* 1. butterfly: w^0
	addf	*ar0,*ar1,r2		; AR' = r2 = AR + BR
	subf	*ar1++,*ar0++,r3	; BR' = r3 = AR - BR
	addf	*ar0,*ar1,r0		; AI' = r0 = AI + BI
	subf	*ar1++,*ar0++,r1	; BI' = r1 = AI - BI
* 2. butterfly: w^0
	addf	*ar0,*ar1,r6		; AR' = r6 = AR + BR
	subf	*ar1++,*ar0++,r7	; BR' = r7 = AR - BR
	addf	*ar0,*ar1,r4		; AI' = r4 = AI + BI
	subf	*ar1++(ir0),*ar0++(ir0),r5	; BI' = r5 = AI - BI
	stf	r2,*ar2++		; (AR' = r2)
||	stf	r3,*ar3++		; (BR' = r3)
	stf	r0,*ar2++		; (AI' = r0)
||	stf	r1,*ar3++		; (BI' = r1)
	stf	r6,*ar2++		; AR' = r6
||	stf	r7,*ar3++		; BR' = r7
	stf	r4,*ar2++(ir0)		; AI' = r4
||	stf	r5,*ar3++(ir0)		; BI' = r5
* 3. butterfly: w^M/4
	addf	*ar0++,*+ar1,r5		; AR' = r5 = AR + BI
	subf	*ar1,*ar0,r4		; AI' = r4 = AI - BR
	addf	*ar1++,*ar0--,r6	; BI' = r6 = AI + BR
	subf	*ar1++,*ar0++,r7	; BR' = r7 = AR - BI
* 4. butterfly: w^M/4
	addf	*+ar1,*++ar0,r3		; AR' = r3 = AR + BI 
	ldf	*-ar7,r1		; r1 = 0 (for inner loop)
||	ldf	*ar1++,r0		; r0 = BR (for inner loop)
	subf	*ar1++(ir0),*ar0++,r2	; BR' = r2 = AR - BI 
	stf	r5,*ar2++		; (AR' = r5)
||	stf	r7,*ar3++		; (BR' = r7)
	stf	r6,*ar3++		; (BI' = r6)

* 5. to M. butterfly: 

	rptb	bf2end	

	ldf	*ar7++,r7		; r7 = COS , ((AI' = r4))
||	stf	r4,*ar2++
	ldf	*ar7++,r6		; r6 = SIN , (BR' = r2)
||	stf	r2,*ar3++
	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r3)
||	stf	r3,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1) 
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++(ir0),r4	; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++(ir0)
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r7,r1		; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++(ir0)
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (AR' = r5)
||	stf	r5,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1) 
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++(ir0),r7,r1	; r1 = BI * COS , (AI' = r4)
||	stf	r4,*ar2++
	addf	*ar0++,r3,r3	 	; r3 = AR + TR , BR' = r2
||	stf	r2,*ar3++

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r3)
||	stf	r3,*ar2++
	subf	r1,r0,r2		; (r2 = TI = r0 - r1) 
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++(ir0),r4	; (r4 = AI - TI , BI' = r3)
||	stf	r3,*ar3++(ir0)
	subf	r0,r5,r3		; r3 = TR = r5 - r0
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2
	mpyf	*ar1++,r6,r1		; r1 = BI * SIN , (AI' = r4)
||	stf	r4,*ar2++(ir0)
	addf	*ar0++,r3,r5		; r5 = AR + TR , BR' = r2
||	stf	r2,*ar3++ 	

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (AR' = r5)
||	stf	r5,*ar2++
	addf	r1,r0,r2		; (r2 = TI = r0 + r1) 
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (r3 = AI + TI)
||	addf	r2,*ar0,r3
	subf	r2,*ar0++,r4		; (r4 = AI - TI , y(L) = BI' = r3)
||	stf	r3,*ar3++
	subf	r0,r5,r3		; r3 = TR = r5 - r0
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , r2 = AR - TR 
||	subf	r3,*ar0,r2
bf2end	mpyf	*ar1++(ir0),r6,r1	; r1 = BI * SIN , r3 = AR + TR
||	addf	*ar0++,r3,r3	

* clear pipeline

	stf	r2,*ar3++		; BR' = r2 , AI' = r4
||	stf	r4,*ar2++
	addf	r1,r0,r2		; r2 = TI = r0 + r1
	addf	r2,*ar0,r3		; r3 = AI + TI , AR' = r3
||	stf	r3,*ar2++		
	subf	r2,*ar0,r4		; r4 = AI - TI , BI' = r3
||	stf	r3,*ar3	
	stf	r4,*ar2			; AI' = r4

*****************************************************************************
* ------------ LAST STAGE WITH INTEGRATED BITREVERSAL --------------------- *
*****************************************************************************

	ldi	@input,ar0	; upper input
	ldi	@output,ar2	; real output !!!
	ldi	@inputp2,ar1    ; lower input
	ldi	@outp1,ar3	; imaginary output !!!
	ldi	@sintp2,ar7	; pointer to twiddle factors
	ldi	@fftsiz,ir0	; bitreversal
	ldi	3,ir1		; group offset
	ldi	@fg4m2,rc

* fill pipeline 

* 1. butterfly: w^0
	addf	*ar0,*ar1,r6		; AR' = r6 = AR + BR
	subf	*ar1++,*ar0++,r7	; BR' = r7 = AR - BR
	subf	*ar1,*ar0,r4		; BI' = r4 = AI - BI
	addf	*ar1++(ir1),*ar0++(ir1),r5	; AI' = r5 = AI + BI
* 2. butterfly: w^M/4
	subf	*+ar1,*ar0,r3		; BR' = r3 = AR - BI
	ldf	*-ar7,r1		; r1 = 0 (for inner loop) 
||	ldf	*ar1++,r0		; r0 = BR (for inner loop)
	addf	*ar1++(ir1),*ar0++,r2	; AR' = r2 = AR + BI 
	stf	r6,*ar2++(ir0)b		; (AR' = r6)
||	stf	r5,*ar3++(ir0)b		; (AI' = r5)
	stf	r7,*ar2++(ir0)b		; (BR' = r7)

* 3. to M. butterfly: 
	rptb	bflend	
* 17 cycles if fftsize<1024 due to the use of internal memory for bitreversal
* 21 cycles if fftsize=1024 due to the use of external memory for bitreversal 
	ldf	*ar7++,r7		; r7 = COS , ((BI' = r4))
||	stf	r4,*ar3++(ir0)b
	ldf	*ar7++,r6		; r6 = SIN , (AR' = r2)
||	stf	r2,*ar2++(ir0)b
	mpyf	*+ar1,r6,r5		; r5 = BI * SIN , (BR' = r3)  
||	stf	r3,*ar2++(ir0)b
	addf	r1,r0,r2		; (r2 = TI = r0 + r1)
	mpyf	*ar1,r7,r0		; r0 = BR * COS , (AI' = r3 = AI - TI)
||	subf	r2,*ar0,r3
	addf	r2,*ar0++(ir1),r4	; (BI' = r4 = AI + TI , AI' = r3)
||	stf	r3,*ar3++(ir0)b
	addf	r0,r5,r3		; r3 = TR = r0 + r5
	mpyf	*ar1++,r6,r0		; r0 = BR * SIN , AR' = r2 = AR + TR
||	addf	r3,*ar0,r2 
	mpyf	*ar1++(ir1),r7,r1	; r1 = BI * COS , (BI' = r4)
||	stf	r4,*ar3++(ir0)b
	subf	r3,*ar0++,r3		; BR' = r3 = AR - TR , AR' = r2
||	stf	r2,*ar2++(ir0)b

	mpyf	*+ar1,r7,r5		; r5 = BI * COS , (BR' = r3)  
||	stf	r3,*ar2++(ir0)b
	subf	r1,r0,r2		; (r2 = TI = r0 - r1)
	mpyf	*ar1,r6,r0		; r0 = BR * SIN , (AI' = r3 = AI - TI)
||	subf	r2,*ar0,r3
	addf	r2,*ar0++(ir1),r4	; (BI' = r4 = AI + TI , AI' = r3)
||	stf	r3,*ar3++(ir0)b
	subf	r0,r5,r3		; r3 = TR = r0 - r5
	mpyf	*ar1++,r7,r0		; r0 = BR * COS , AR' = r2 = AR + TR 
||	addf	r3,*ar0,r2 
bflend	mpyf	*ar1++(ir1),r6,r1	; r1 = BI * SIN , BR' = r3 = AR - TR
||	subf	r3,*ar0++,r3	

* clear pipeline

	stf	r2,*ar2++(ir0)b		; AR' = r2 , (BI' = r4)
||	stf	r4,*ar3++(ir0)b
	addf	r1,r0,r2		; r2 = TI = r0 + r1
	subf	r2,*ar0,r3		; AI' = r3 = AI - TI , BR' = r3
||	stf	r3,*ar2
	addf	r2,*ar0,r4		; BI' = r4 = AI + TI , AI' = r3
||	stf	r3,*ar3++(ir0)b			
	stf	r4,*ar3			; BI' = r4

*****************************************************************************
*------------- END OF FFT --------------------------------------------------*
*****************************************************************************

end:	nop
	nop
	nop
	nop

self	br	self
	.end

